@@ -1,14 +1,18 @@
-from typing import Dict, Iterable, List, Optional, Set, Tuple, TypeVar, Union
+from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple, TypeVar, Union
 
-from google.cloud.bigquery import Client, Row, SchemaField
+from google.cloud.bigquery import Row, SchemaField
 from google.cloud.bigquery.table import RowIterator
 
-from bigquery_frame.auth import get_bq_client
 from bigquery_frame.column import Column, StringOrColumn, cols_to_str
-from bigquery_frame.has_bigquery_client import HasBigQueryClient
+from bigquery_frame.conf import ELEMENT_COL_NAME, REPETITION_MARKER, STRUCT_SEPARATOR
+from bigquery_frame.nested import resolve_nested_columns
 from bigquery_frame.printing import print_results
+from bigquery_frame.transformations_impl.flatten_schema import flatten_schema
 from bigquery_frame.utils import assert_true, indent, quote, str_to_col, strip_margin
 
+if TYPE_CHECKING:
+    from bigquery_frame.bigquery_builder import BigQueryBuilder
+
 A = TypeVar("A")
 B = TypeVar("B")
 
@@ -57,7 +61,8 @@ def schema_to_simple_string(schema: List[SchemaField]):
 
     Example:
 
-    >>> bq = BigQueryBuilder(get_bq_client())
+    >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+    >>> bq = BigQueryBuilder()
     >>> df = bq.sql('SELECT 1 as id, STRUCT(1 as a, [STRUCT(2 as c, 3 as d)] as b, [4, 5] as e) as s')
     >>> print(df.schema)  # noqa: E501
     [SchemaField('id', 'INTEGER', 'NULLABLE', None, (), None), SchemaField('s', 'RECORD', 'NULLABLE', None, (SchemaField('a', 'INTEGER', 'NULLABLE', None, (), None), SchemaField('b', 'RECORD', 'REPEATED', None, (SchemaField('c', 'INTEGER', 'NULLABLE', None, (), None), SchemaField('d', 'INTEGER', 'NULLABLE', None, (), None)), None), SchemaField('e', 'INTEGER', 'REPEATED', None, (), None)), None)]
@@ -116,67 +121,6 @@ def _dedup_key_value_list(items: List[Tuple[A, B]]) -> List[Tuple[A, B]]:
     return list({k: v for k, v in items}.items())
 
 
-class BigQueryBuilder(HasBigQueryClient):
-    def __init__(self, client: Client):
-        super().__init__(client)
-        self._alias_count = 0
-        self._temp_table_count = 0
-        self._views: Dict[str, "DataFrame"] = {}
-        self._temp_tables: Set[str] = set()
-
-    def table(self, full_table_name: str) -> "DataFrame":
-        """Returns the specified table as a :class:`DataFrame`."""
-        query = f"""SELECT * FROM {quote(full_table_name)}"""
-        return DataFrame(query, alias=None, bigquery=self)
-
-    def sql(self, sql_query) -> "DataFrame":
-        """Returns a :class:`DataFrame` representing the result of the given query."""
-        return DataFrame(sql_query, None, self)
-
-    def _registerDataFrameAsTempView(self, df: "DataFrame", alias: str) -> None:
-        # self._check_alias(alias, [])
-        self._views[alias] = df
-
-    def _registerDataFrameAsTempTable(self, df: "DataFrame", alias: Optional[str] = None) -> "DataFrame":
-        if alias is None:
-            alias = self._get_temp_table_alias()
-        query = f"CREATE OR REPLACE TEMP TABLE {quote(alias)} AS \n" + df.compile()
-        self._execute_query(query)
-        return self.table(alias)
-
-    def _compile_views(self) -> List[str]:
-        return [
-            strip_margin(
-                f"""{quote(alias)} AS (
-                |{indent(df._compile_with_deps(), 2)}
-                |)"""
-            )
-            for alias, df in self._views.items()
-        ]
-
-    def _get_alias(self) -> str:
-        self._alias_count += 1
-        return DEFAULT_ALIAS_NAME.format(num=self._alias_count)
-
-    def _get_temp_table_alias(self) -> str:
-        self._temp_table_count += 1
-        return DEFAULT_TEMP_TABLE_NAME.format(num=self._temp_table_count)
-
-    def _check_alias(self, new_alias, deps: List[Tuple[str, "DataFrame"]]) -> None:
-        """Checks that the alias follows BigQuery constraints, such as:
-
-        - BigQuery does not allow having two CTEs with the same name in a query.
-
-        :param new_alias:
-        :param deps:
-        :return: None
-        :raises: an Exception if something that does not comply with BigQuery's rules is found.
-        """
-        collisions = [alias for alias, df in list(self._views.items()) + deps if alias == new_alias]
-        if len(collisions) > 0:
-            raise Exception(f"Duplicate alias {new_alias}")
-
-
 class DataFrame:
 
     _deps: List[Tuple[str, "DataFrame"]]
@@ -186,7 +130,7 @@ class DataFrame:
         self,
         query: str,
         alias: Optional[str],
-        bigquery: BigQueryBuilder,
+        bigquery: "BigQueryBuilder",
         deps: Optional[List["DataFrame"]] = None,
     ):
         self.query = query
@@ -199,7 +143,7 @@ class DataFrame:
         else:
             bigquery._check_alias(alias, self._deps)
         self._alias = alias
-        self.bigquery: BigQueryBuilder = bigquery
+        self.bigquery: "BigQueryBuilder" = bigquery
         self._schema = None
 
     def __repr__(self):
@@ -260,23 +204,29 @@ class DataFrame:
 
     def _compute_schema(self):
         df = self.limit(0)
-        return df.bigquery._execute_query(df.compile()).schema
+        return df.bigquery._execute_query(df.compile(), use_query_cache=False).schema
 
-    def _compile_deps(self) -> List[str]:
-        return [
-            strip_margin(
+    def _compile_deps(self) -> Dict[str, str]:
+        return {
+            alias: strip_margin(
                 f"""{quote(alias)} AS (
                 |{indent(cte.query, 2)}
                 |)"""
             )
             for (alias, cte) in self._deps
-        ]
+        }
 
-    def _compile_with_ctes(self, ctes: List[str]) -> str:
+    def _compile_with_ctes(self, ctes: Dict[str, str]) -> str:
         if len(ctes) > 0:
-            return "WITH " + "\n, ".join(ctes) + "\n" + self.query
+            query = "WITH " + "\n, ".join(ctes.values()) + "\n" + self.query
         else:
-            return self.query
+            query = self.query
+        deps_replacements = {
+            alias[1:-1]: DEFAULT_ALIAS_NAME.format(num=i + 1)
+            for (i, (alias, _)) in enumerate(ctes.items())
+            if alias[0] == "{" and alias[-1] == "}"
+        }
+        return query.format(**deps_replacements)
 
     def _compile_with_deps(self) -> str:
         ctes = self._compile_deps()
@@ -313,13 +263,14 @@ class DataFrame:
 
     def compile(self) -> str:
         """Returns the sql query that will be executed to materialize this :class:`DataFrame`"""
-        ctes = self.bigquery._compile_views() + self._compile_deps()
+        ctes = {**self.bigquery._compile_views(), **self._compile_deps()}
         return self._compile_with_ctes(ctes)
 
-    def createOrReplaceTempTable(self, alias: str) -> "DataFrame":
+    def createOrReplaceTempTable(self, alias: str) -> None:
         """Creates or replace a persisted temporary table.
 
-        >>> bq = BigQueryBuilder(get_bq_client())
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+            >>> bq = BigQueryBuilder()
         >>> df = bq.sql("SELECT 1 as id")
         >>> df.createOrReplaceTempTable("temp_table")
         >>> bq.sql("SELECT * FROM temp_table").show()
@@ -336,7 +287,8 @@ class DataFrame:
     def createOrReplaceTempView(self, name: str) -> None:
         """Creates or replaces a local temporary view with this :class:`DataFrame`.
 
-        >>> bq = BigQueryBuilder(get_bq_client())
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+            >>> bq = BigQueryBuilder()
         >>> df = bq.sql("SELECT 1 as id")
         >>> df.createOrReplaceTempView("temp_view")
         >>> bq.sql("SELECT * FROM temp_view").show()
@@ -635,7 +587,8 @@ class DataFrame:
 
         Examples:
 
-        >>> bq = BigQueryBuilder(get_bq_client())
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+            >>> bq = BigQueryBuilder()
         >>> df = bq.sql('''SELECT 1 as id, STRUCT(1 as a, [STRUCT(1 as c)] as b) as s''')
         >>> df.printSchema()
         root
@@ -654,7 +607,8 @@ class DataFrame:
 
         Examples:
 
-        >>> bq = BigQueryBuilder(get_bq_client())
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+            >>> bq = BigQueryBuilder()
         >>> from bigquery_frame import functions as f
         >>> df = bq.sql('''SELECT 1 as a''').select(
         ...   'a', f.col('a') + f.lit(1).alias('b')).withColumn('c', f.expr('a + b')
@@ -688,13 +642,93 @@ class DataFrame:
         )
         return self._apply_query(query)
 
+    def select_nested_columns(self, columns: Dict[str, StringOrColumn]) -> "DataFrame":
+        """Projects a set of expressions and returns a new :class:`DataFrame`.
+        Unlike the :func:`select` method, this method works on repeated elements
+        and records (arrays and arrays of struct).
+
+        The syntax for column names works as follow:
+        - "." is the separator for struct elements
+        - "!" must be appended at the end of fields that are repeated
+        - Corresponding column expressions must use the complete field names for non-repeated structs
+        - Corresponding column expressions must use short field names for repeated structs
+        - Corresponding column expressions must use the name `_` for array elements
+
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+            >>> bq = BigQueryBuilder()
+        >>> from bigquery_frame import functions as f
+        >>> df = bq.sql('''
+        ...  SELECT * FROM UNNEST([
+        ...    STRUCT(1 as id, STRUCT(1 as a, 2 as b) as s)
+        ...  ])
+        ... ''')
+        >>> df.show()
+        +----+------------------+
+        | id |                s |
+        +----+------------------+
+        |  1 | {'a': 1, 'b': 2} |
+        +----+------------------+
+        >>> df.select_nested_columns({
+        ...     "id": "id",
+        ...     "s.c": f.col("s.a") + f.col("s.b")
+        ... }).show()
+        +----+----------+
+        | id |        s |
+        +----+----------+
+        |  1 | {'c': 3} |
+        +----+----------+
+
+        >>> from bigquery_frame import functions as f
+        >>> bq = BigQueryBuilder()
+        >>> df = bq.sql('''
+        ...  SELECT * FROM UNNEST([
+        ...    STRUCT(1 as id, [STRUCT(1 as a, 2 as b), STRUCT(3 as a, 4 as b)] as s)
+        ...  ])
+        ... ''')
+        >>> df.show()
+        +----+--------------------------------------+
+        | id |                                    s |
+        +----+--------------------------------------+
+        |  1 | [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}] |
+        +----+--------------------------------------+
+        >>> df.select_nested_columns({"s!.c": f.col("a") + f.col("b")}).show()
+        +----------------------+
+        |                    s |
+        +----------------------+
+        | [{'c': 3}, {'c': 7}] |
+        +----------------------+
+
+        >>> df = bq.sql('''
+        ...  SELECT * FROM UNNEST([
+        ...    STRUCT(1 as id, [STRUCT([1, 2, 3] as e)] as s)
+        ...  ])
+        ... ''')
+        >>> df.show()
+        +----+--------------------+
+        | id |                  s |
+        +----+--------------------+
+        |  1 | [{'e': [1, 2, 3]}] |
+        +----+--------------------+
+        >>> df.select_nested_columns({"s!.e!": f.col("_").cast("FLOAT64")}).show()
+        +--------------------------+
+        |                        s |
+        +--------------------------+
+        | [{'e': [1.0, 2.0, 3.0]}] |
+        +--------------------------+
+
+        :param columns: a Dict[column_alias, column_expression] of columns to select
+        :return:
+        """
+        return self.select(*resolve_nested_columns(columns))
+
     def show(self, n: int = 20, format_args=None) -> None:
         """Prints the first ``n`` rows to the console. This uses the awesome Python library called `tabulate
         <https://pythonrepo.com/repo/astanin-python-tabulate-python-generating-and-working-with-logs>`_.
 
         Formating options may be set using `format_args`.
 
-        >>> bq = BigQueryBuilder(get_bq_client())
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+            >>> bq = BigQueryBuilder()
         >>> df = bq.sql('''SELECT 1 as id, STRUCT(1 as a, [STRUCT(1 as c)] as b) as s''')
         >>> df.show()
         +----+---------------------------+
@@ -825,7 +859,8 @@ class DataFrame:
         The difference between this function and :func:`union` is that this function
         resolves columns by name (not by position):
 
-        >>> bq = BigQueryBuilder(get_bq_client())
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+            >>> bq = BigQueryBuilder()
         >>> df1 = bq.sql('''SELECT 1 as col0, 2 as col1, 3 as col2''')
         >>> df2 = bq.sql('''SELECT 4 as col1, 5 as col2, 6 as col0''')
         >>> df1.union(df2).show()
@@ -925,6 +960,129 @@ class DataFrame:
             query = f"SELECT *, {col_expr} AS {col_name} FROM {quote(self._alias)}"
         return self._apply_query(query)
 
+    def with_nested_columns(self, columns: Dict[str, StringOrColumn]) -> "DataFrame":
+        """Returns a new :class:`DataFrame` by adding or replacing (when they already exist) columns.
+
+        Unlike the :func:`withColumn` method, this method works on repeated elements
+        and records (arrays and arrays of struct).
+
+        The syntax for column names works as follow:
+        - "." is the separator for struct elements
+        - "!" must be appended at the end of fields that are repeated
+        - Corresponding column expressions must use the complete field names for non-repeated structs
+        - Corresponding column expressions must use short field names for repeated structs
+        - Corresponding column expressions must use the name `_` for array elements
+
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+            >>> bq = BigQueryBuilder()
+        >>> from bigquery_frame import functions as f
+        >>> df = bq.sql('''
+        ...  SELECT * FROM UNNEST([
+        ...    STRUCT(1 as id, STRUCT(1 as a, 2 as b) as s)
+        ...  ])
+        ... ''')
+        >>> df.show()
+        +----+------------------+
+        | id |                s |
+        +----+------------------+
+        |  1 | {'a': 1, 'b': 2} |
+        +----+------------------+
+        >>> df.with_nested_columns({
+        ...     "id": "id",
+        ...     "s.c": f.col("s.a") + f.col("s.b")
+        ... }).show()
+        +----+--------------------------+
+        | id |                        s |
+        +----+--------------------------+
+        |  1 | {'a': 1, 'b': 2, 'c': 3} |
+        +----+--------------------------+
+
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+            >>> from bigquery_frame import functions as f
+        >>> bq = BigQueryBuilder()
+        >>> df = bq.sql('''
+        ...  SELECT * FROM UNNEST([
+        ...    STRUCT(1 as id, [STRUCT(1 as a, 2 as b), STRUCT(3 as a, 4 as b)] as s)
+        ...  ])
+        ... ''')
+        >>> df.show()
+        +----+--------------------------------------+
+        | id |                                    s |
+        +----+--------------------------------------+
+        |  1 | [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}] |
+        +----+--------------------------------------+
+        >>> df.with_nested_columns({"s!.c": f.col("a") + f.col("b")}).show()
+        +----+------------------------------------------------------+
+        | id |                                                    s |
+        +----+------------------------------------------------------+
+        |  1 | [{'a': 1, 'b': 2, 'c': 3}, {'a': 3, 'b': 4, 'c': 7}] |
+        +----+------------------------------------------------------+
+
+        >>> df = bq.sql('''
+        ...  SELECT * FROM UNNEST([
+        ...    STRUCT(1 as id, [STRUCT([1, 2, 3] as e)] as s)
+        ...  ])
+        ... ''')
+        >>> df.show()
+        +----+--------------------+
+        | id |                  s |
+        +----+--------------------+
+        |  1 | [{'e': [1, 2, 3]}] |
+        +----+--------------------+
+        >>> df.with_nested_columns({"s!.e!": f.col("_").cast("FLOAT64")}).show()
+        +----+--------------------------+
+        | id |                        s |
+        +----+--------------------------+
+        |  1 | [{'e': [1.0, 2.0, 3.0]}] |
+        +----+--------------------------+
+
+        :param columns: a Dict[column_alias, column_expression] of columns to select
+        :return:
+        """
+        schema_flat = flatten_schema(
+            self.schema, explode=True, struct_separator=STRUCT_SEPARATOR, repetition_marker=REPETITION_MARKER
+        )
+
+        def get_col_short_name(col: str):
+            if col[-1] == REPETITION_MARKER:
+                return ELEMENT_COL_NAME
+            else:
+                return col.split(REPETITION_MARKER + STRUCT_SEPARATOR)[-1]
+
+        default_columns = {field.name: get_col_short_name(field.name) for field in schema_flat}
+        columns = {**default_columns, **columns}
+        return self.select(*resolve_nested_columns(columns))
+
+    @property
+    def write(self):
+        """Interface for saving the content of the :class:`DataFrame` out into external storage.
+
+        Examples
+        --------
+        >>> from bigquery_frame.bigquery_builder import BigQueryBuilder
+        >>> from bigquery_frame.auth import get_bq_client
+        >>> from bigquery_frame.dataframe_writer import __setup_test_dataset, __teardown_test_dataset
+        >>> client = get_bq_client()
+        >>> bq = BigQueryBuilder(client)
+        >>> test_dataset = __setup_test_dataset(client)
+        >>> df = bq.sql("SELECT 1 as a")
+
+        >>> df.write.mode('overwrite').save(f"{test_dataset.dataset_id}.my_table")
+        >>> df.write.mode('append').save(f"{test_dataset.dataset_id}.my_table")
+        >>> bq.table(f"{test_dataset.dataset_id}.my_table").show()
+        +---+
+        | a |
+        +---+
+        | 1 |
+        | 1 |
+        +---+
+        >>> __teardown_test_dataset(client, test_dataset)
+
+        """
+        from bigquery_frame.dataframe_writer import DataframeWriter
+
+        return DataframeWriter(self)
+
     orderBy = sort
 
     unionAll = union
@@ -933,6 +1091,9 @@ class DataFrame:
 
 
 def __get_test_df() -> DataFrame:
+    from bigquery_frame.auth import get_bq_client
+    from bigquery_frame.bigquery_builder import BigQueryBuilder
+
     bq = BigQueryBuilder(get_bq_client())
     query = """
         SELECT
@@ -947,6 +1108,9 @@ def __get_test_df() -> DataFrame:
 
 
 def __get_test_dfs() -> Tuple[DataFrame, ...]:
+    from bigquery_frame.auth import get_bq_client
+    from bigquery_frame.bigquery_builder import BigQueryBuilder
+
     bq = BigQueryBuilder(get_bq_client())
     df1 = bq.sql(
         strip_margin(
@@ -983,6 +1147,9 @@ def __get_test_dfs() -> Tuple[DataFrame, ...]:
 
 
 def __get_test_df_2() -> DataFrame:
+    from bigquery_frame.auth import get_bq_client
+    from bigquery_frame.bigquery_builder import BigQueryBuilder
+
     bq = BigQueryBuilder(get_bq_client())
     query = """
         SELECT
