@@ -1,11 +1,13 @@
-from typing import List, Optional, Set, Union
+from typing import Iterable, List, Optional, Set, Union
 
 from bigquery_frame import BigQueryBuilder
 from bigquery_frame.auth import get_bq_client
 from bigquery_frame.column import (
+    ArrayColumn,
     Column,
     LitOrColumn,
     StringOrColumn,
+    WhenColumn,
     cols_to_str,
     literal_col,
 )
@@ -182,7 +184,7 @@ def concat(*cols: StringOrColumn) -> Column:
 
     Examples
     --------
-    >>> bq = BigQueryBuilder(get_bq_client())
+    >>> bq = BigQueryBuilder()
     >>> df = bq.sql("SELECT 'abcd' as s, '123' as d")
     >>> from bigquery_frame import functions as f
     >>> df.select(f.concat(df['s'], df['d']).alias('s')).show()
@@ -345,7 +347,7 @@ def length(col: StringOrColumn) -> Column:
 
     Examples
     --------
-    >>> bq = BigQueryBuilder(get_bq_client())
+    >>> bq = BigQueryBuilder()
     >>> from bigquery_frame import functions as f
     >>> bq.sql("SELECT 'ABC ' as a").select(f.length('a').alias('length')).show()
     +--------+
@@ -419,7 +421,7 @@ def replace(original_value: StringOrColumn, from_value: LitOrColumn, replace_val
     """Replaces all occurrences of `from_value` with `to_value` in `original_value`.
     If `from_value` is empty, no replacement is made.
 
-    >>> bq = BigQueryBuilder(get_bq_client())
+    >>> bq = BigQueryBuilder()
     >>> df = bq.sql("SELECT 'a.b.c.d' as s, '.' as dot, '/' as slash")
     >>> from bigquery_frame import functions as f
     >>> df.select(f.replace('s', ".", "/").alias('s')).show()
@@ -460,13 +462,15 @@ def replace(original_value: StringOrColumn, from_value: LitOrColumn, replace_val
     return Column(f"REPLACE({original_value.expr}, {from_value.expr}, {replace_value.expr})")
 
 
-def sort_array(col: StringOrColumn, asc: bool = True) -> Column:
+def sort_array(array: StringOrColumn, sort_cols: Union[Column, List[Column]]) -> ArrayColumn:
     """Collection function: sorts the input array in ascending or descending order according to the natural ordering
     of the array elements. Unlike in Spark, arrays cannot contain NULL element when they are serialized.
 
+    Array elements are implicitely called `_` and can be referred to as such.
+
     Examples
     --------
-    >>> bq = BigQueryBuilder(get_bq_client())
+    >>> bq = BigQueryBuilder()
     >>> df = bq.sql('''
     ...     SELECT data FROM UNNEST ([
     ...         STRUCT([2, 1, 3] as data),
@@ -474,7 +478,7 @@ def sort_array(col: StringOrColumn, asc: bool = True) -> Column:
     ...         STRUCT([] as data)])
     ... ''')
     >>> from bigquery_frame import functions as f
-    >>> df.select(f.sort_array(df['data']).alias('r')).show()
+    >>> df.select(f.sort_array(df['data'], f.col("_")).alias('r')).show()
     +-----------+
     |         r |
     +-----------+
@@ -482,7 +486,7 @@ def sort_array(col: StringOrColumn, asc: bool = True) -> Column:
     |       [1] |
     |        [] |
     +-----------+
-    >>> df.select(f.sort_array(df['data'], asc=False).alias('r')).show()
+    >>> df.select(f.sort_array(df['data'], f.col("_").desc()).alias('r')).show()
     +-----------+
     |         r |
     +-----------+
@@ -491,15 +495,16 @@ def sort_array(col: StringOrColumn, asc: bool = True) -> Column:
     |        [] |
     +-----------+
 
-    :param col: `Column` or str name of column
-    :param asc: bool, optional
+    :param array: `Column` or str name of column of type ARRAY
+    :param sort_cols: `Column` or str, or list of `Column` or str specifying how the array must be sorted.
+         Array elements can be referred to as `_`.
     :return:
     """
-    col = str_to_col(col)
-    desc_str = ""
-    if not asc:
-        desc_str = " DESC"
-    return Column(f"ARRAY(SELECT elem FROM UNNEST({col.expr}) elem ORDER BY elem{desc_str})")
+    if not isinstance(sort_cols, Iterable):
+        sort_cols = [sort_cols]
+    if isinstance(array, ArrayColumn):
+        return array._copy(sort_cols=sort_cols)
+    return ArrayColumn(array, sort_cols=sort_cols)
 
 
 def substring(col: StringOrColumn, pos: LitOrColumn, len: Optional[LitOrColumn] = None) -> Column:
@@ -512,7 +517,7 @@ def substring(col: StringOrColumn, pos: LitOrColumn, len: Optional[LitOrColumn]
 
     Examples
     --------
-    >>> bq = BigQueryBuilder(get_bq_client())
+    >>> bq = BigQueryBuilder()
     >>> df = bq.sql("SELECT 'abcd' as s")
     >>> from bigquery_frame import functions as f
     >>> df.select(f.substring(df['s'], 1, 2).alias('s')).show()
@@ -604,7 +609,125 @@ def struct(*cols: Union[StringOrColumn, List[StringOrColumn], Set[StringOrColumn
     return Column(f"STRUCT({cols_to_str(cols)})")
 
 
-def when(condition: Column, value: Column) -> Column:
+def to_base32(col: StringOrColumn) -> Column:
+    """Converts a sequence of BYTES into a base32-encoded STRING.
+    To convert a base32-encoded STRING into BYTES, use :func:`from_base32`.
+
+    Examples
+    --------
+    >>> bq = BigQueryBuilder()
+    >>> from bigquery_frame import functions as f
+    >>> df = bq.sql(r"SELECT b'abcde\\xFF' as b")
+    >>> df.select(f.to_base32('b').alias('base32_string')).show()
+    +------------------+
+    |    base32_string |
+    +------------------+
+    | MFRGGZDF74====== |
+    +------------------+
+
+    """
+    return _invoke_function_over_column("TO_BASE32", col)
+
+
+def to_base64(col: StringOrColumn) -> Column:
+    """Converts a sequence of BYTES into a base64-encoded STRING.
+    To convert a base64-encoded STRING into BYTES, use :func:`from_base64`.
+
+    There are several base64 encodings in common use that vary in exactly which alphabet of 65 ASCII characters
+    are used to encode the 64 digits and padding.
+    See `RFC 4648 <https://tools.ietf.org/html/rfc4648#section-4>`_ for details.
+    This function adds padding and uses the alphabet [A-Za-z0-9+/=].
+
+    Examples
+    --------
+    >>> bq = BigQueryBuilder()
+    >>> from bigquery_frame import functions as f
+    >>> df = bq.sql(r"SELECT b'\\377\\340' as b")
+    >>> df.select(f.to_base64('b').alias('base64_string')).show()
+    +---------------+
+    | base64_string |
+    +---------------+
+    |          /+A= |
+    +---------------+
+
+    To work with an encoding using a different base64 alphabet, you might need to compose TO_BASE64 with the
+    REPLACE function. For instance, the base64url url-safe and filename-safe encoding commonly used in web programming
+    uses -_= as the last characters rather than +/=. To encode a base64url-encoded string,
+    replace + and / with - and _ respectively.
+
+    >>> df.select(f.replace(f.replace(f.to_base64('b'), '+', '-'), '/', '_').alias('websafe_base64')).show()
+    +----------------+
+    | websafe_base64 |
+    +----------------+
+    |           _-A= |
+    +----------------+
+
+    """
+    return _invoke_function_over_column("TO_BASE64", col)
+
+
+def transform(array: StringOrColumn, transform_col: Column) -> Column:
+    """Returns an array of elements after applying a transformation to each element in the input array.
+
+    The elements of the array are implicitely called `_`, and the `transform_col` can thus be any Column expression
+    referencing `_`.
+    For arrays of structs (a.k.a. repeated records), the `transform_col` may also directly
+    refer to the fields of the struct.
+
+    Examples
+    --------
+    >>> from bigquery_frame import functions as f
+    >>> bq = BigQueryBuilder()
+    >>> df = bq.sql("SELECT 1 as key, [1, 2, 3, 4] as values")
+    >>> df.select(transform("values", f.col("_") * f.lit(2)).alias("doubled")).show()
+    +--------------+
+    |      doubled |
+    +--------------+
+    | [2, 4, 6, 8] |
+    +--------------+
+
+    >>> def alternate(x, i):
+    ...     return when(i % 2 == 0, x).otherwise(-x)
+    >>> x = f.col("_")
+    >>> df.select(transform("values", f.when(x % 2 == 0, x).otherwise(-x)).alias("alternated")).show()
+    +----------------+
+    |     alternated |
+    +----------------+
+    | [-1, 2, -3, 4] |
+    +----------------+
+
+    >>> df = bq.sql("SELECT 1 as key, [STRUCT(1 as a, 2 as b), STRUCT(3 as a, 4 as b)] as s")
+    >>> df.show()
+    +-----+--------------------------------------+
+    | key |                                    s |
+    +-----+--------------------------------------+
+    |   1 | [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}] |
+    +-----+--------------------------------------+
+    >>> df.select(transform("s",
+    ...     f.struct(
+    ...         f.expr("a * 2").alias("double_a"),
+    ...         f.expr("b * 3").alias("triple_b")
+    ...     )
+    ... ).alias("s")).show()
+    +-------------------------------------------------------------------+
+    |                                                                 s |
+    +-------------------------------------------------------------------+
+    | [{'double_a': 2, 'triple_b': 6}, {'double_a': 6, 'triple_b': 12}] |
+    +-------------------------------------------------------------------+
+
+    :param array: `Column` or str name of column of type ARRAY
+    :param transform_col: `Column` or str specifying the transformation to apply.
+        Array elements can be referred to as `_`. If the elements are structs, their fields can be referred to directly.
+    :return:
+    """
+    if isinstance(array, ArrayColumn):
+        return array._copy(transform_col=transform_col)
+    else:
+        array = str_to_col(array)
+        return ArrayColumn(array, transform_col=transform_col)
+
+
+def when(condition: Column, value: Column) -> WhenColumn:
     """Evaluates a list of conditions and returns one of multiple possible result expressions.
     If :func:`Column.otherwise` is not invoked, None is returned for unmatched conditions.
 
@@ -625,9 +748,7 @@ def when(condition: Column, value: Column) -> Column:
     :param value: a :class:`Column` expression.
     :return:
     """
-    c = Column("")
-    c._when_condition = [(condition, value)]
-    return c
+    return WhenColumn([(condition, value)])
 
 
 def _get_test_df_1() -> DataFrame:
