@@ -8,7 +8,7 @@ from bigquery_frame.auth import get_bq_client
 from bigquery_frame.column import StringOrColumn, cols_to_str
 from bigquery_frame.dataframe import strip_margin
 from bigquery_frame.transformations_impl import analyze_aggs
-from bigquery_frame.transformations_impl.flatten import flatten_schema
+from bigquery_frame.transformations_impl.flatten_schema import flatten_schema
 from bigquery_frame.transformations_impl.union_dataframes import union_dataframes
 from bigquery_frame.utils import quote
 
@@ -16,7 +16,7 @@ from bigquery_frame.utils import quote
 def _unnest_column(df: DataFrame, col: str, extra_cols: Optional[List[str]] = None):
     """Recursively unnest a :class:`DataFrame`'s column
 
-    >>> bq = BigQueryBuilder(get_bq_client())
+    >>> bq = BigQueryBuilder()
     >>> df = bq.sql('SELECT 1 as id, '
     ... '[STRUCT(2 as a, [STRUCT(3 as c, 4 as d)] as b, [5, 6] as e)] as s1, STRUCT(7 as f) as s2')
     >>> [col.name for col in flatten_schema(df.schema, explode=True)]
@@ -133,11 +133,12 @@ def _analyze_column(df: DataFrame, schema_field: SchemaField, col_num: int, grou
         else:
             col = col.split(".")[-1]
         group_alias = quote("__group1__")
-        group_by_1 = [df[col] for col in group_by]
+        group_by_1 = [df[col].alias(col.replace(".", "_")) for col in group_by]
+        group_by_2 = [f.expr(f"{group_alias}.{quote(col.replace('.', '_'))}") for col in group_by]
         group_select_1 = [f.struct(*group_by_1).alias(group_alias)] if len(group_by) > 0 else []
         df = _unnest_column(df, schema_field.name, extra_cols=group_select_1)
-
-    group_by_2 = [f.expr(f"{group_alias}.{quote(col)}") for col in group_by]
+    else:
+        group_by_2 = [f.expr(f"{group_alias}.{quote(col)}") for col in group_by]
     group_select_2 = [f.struct(*group_by_2).alias(quote("group"))] if len(group_by) > 0 else []
 
     res = _select_group_by(df, *group_select_2, *[agg(col, schema_field, col_num) for agg in aggs], group_by=group_by_2)
