@@ -6,8 +6,7 @@ def sort_columns(df: DataFrame) -> DataFrame:
 
     Examples:
     >>> from bigquery_frame import BigQueryBuilder
-    >>> from bigquery_frame.auth import get_bq_client
-    >>> bq = BigQueryBuilder(get_bq_client())
+    >>> bq = BigQueryBuilder()
     >>> df = bq.sql('''SELECT 1 as b, 1 as a, 1 as c''')
     >>> df.printSchema()
     root
